﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Markup;
using System.Windows.Media;
using System.Windows.Shapes;

namespace TestGridViewColumEx
{
    //[ContentProperty("Header")]
    //[StyleTypedProperty(Property = "HeaderContainerStyle", StyleTargetType = typeof(GridViewColumnHeaderEx))]
    //[StyleTypedProperty(Property = "HeaderContainerStyle", StyleTargetType = typeof(GridViewColumnHeader))]
    public class GridViewColumnEx : GridViewColumn
    {
        #region constructor & propertychanged

        static GridViewColumnEx()
        {
            HeaderProperty.OverrideMetadata(typeof(GridViewColumnEx),
                new FrameworkPropertyMetadata(typeof(GridViewColumnEx),
                (ss, ee) =>
                {
                    if (ee.NewValue is GridViewColumnHeaderEx)
                    {
                    }
                    else
                    {
                        var gvc = ss as GridViewColumnEx;
                        var gvch = new GridViewColumnHeaderEx()
                        {
                            Content = ee.NewValue is string ? " " + ee.NewValue + " " : ee.NewValue,
                        };
                        // set all binding
                        SetBinding(gvc, gvch, "CanSort", GridViewColumnHeaderEx.CanSortProperty);
                        SetBinding(gvc, gvch, "CanFilter", GridViewColumnHeaderEx.CanFilterProperty);
                        SetBinding(gvc, gvch, "CanHide", GridViewColumnHeaderEx.CanHideProperty);
                        SetBinding(gvc, gvch, "CanMove", GridViewColumnHeaderEx.CanMoveProperty);
                        SetBinding(gvc, gvch, "CanResize", GridViewColumnHeaderEx.CanResizeProperty);
                        // tooltip
                        SetBinding(gvc, gvch, "HideToolTip", GridViewColumnHeaderEx.HideToolTipProperty);
                        SetBinding(gvc, gvch, "FilterToolTip", GridViewColumnHeaderEx.FilterToolTipProperty);
                        // other
                        SetBinding(gvc, gvch, "HorizontalContentAlignment", GridViewColumnHeaderEx.HorizontalContentAlignmentProperty);
                        SetBinding(gvc, gvch, "DisableMoveColor", GridViewColumnHeaderEx.DisableMoveColorProperty);
                        SetBinding(gvc, gvch, "FilterIndicatorColor", GridViewColumnHeaderEx.FilterIndicatorColorProperty);
                        SetBinding(gvc, gvch, "Profil", GridViewColumnHeaderEx.ProfilProperty);
                        SetBinding(gvc, gvch, "ColumnBindingName", GridViewColumnHeaderEx.ColumnBindingNameProperty);
                        SetBinding(gvc, gvch, "FilterEditor", GridViewColumnHeaderEx.FilterEditorProperty);
                        SetBinding(gvc, gvch, "MinColumnWidth", GridViewColumnHeaderEx.MinColumnWidthProperty);
                        // common
                        typeof(GridViewColumnHeader).SetDependencyPropertyKeyValue("ColumnPropertyKey", gvch, gvc);
                        // change header
                        gvc.Header = gvch;
                    }
                    //- local func -
                    void SetBinding(object source, FrameworkElement destination, string property, DependencyProperty dp)
                    {
                        Binding binding = new Binding(property);
                        binding.Source = source;
                        destination.SetBinding(dp, binding);
                    }
                }));

            PropertyChangedCallback pcc = (ss, ee) =>
            {
                if (ss is GridViewColumnEx gvc)
                {
                    if ((gvc.CellTemplate != null || gvc.CellTemplateSelector != null) &&
                            gvc.DisplayMemberBinding != null)
                        gvc.DisplayMemberBinding = null;
                }
            };

            CellTemplateProperty.OverrideMetadata(typeof(GridViewColumnEx),
                new FrameworkPropertyMetadata(null, pcc));
            CellTemplateSelectorProperty.OverrideMetadata(typeof(GridViewColumnEx),
                new FrameworkPropertyMetadata(null, pcc));
        }

        public GridViewColumnEx() { }

        protected override void OnPropertyChanged(PropertyChangedEventArgs e)
        {
            base.OnPropertyChanged(e);
            if (e?.PropertyName?.Equals("DisplayMemberBinding") ?? false)
            {
                if (ColumnBindingName == null && DisplayMemberBinding is Binding b)
                {
                    ColumnBindingName = b.Path.Path;
                }
            }
        }

        #endregion

        #region DP can

        #region CanSort (DP SHORT) [false]
        public bool CanSort { get { return (bool)GetValue(CanSortProperty); } set { SetValue(CanSortProperty, value); } }
        public static readonly DependencyProperty CanSortProperty = DependencyProperty.Register(
            "CanSort", typeof(bool), typeof(GridViewColumnEx), new PropertyMetadata(false));
        #endregion

        #region CanFilter (DP SHORT) [false]
        public bool CanFilter { get { return (bool)GetValue(CanFilterProperty); } set { SetValue(CanFilterProperty, value); } }
        public static readonly DependencyProperty CanFilterProperty = DependencyProperty.Register(
            "CanFilter", typeof(bool), typeof(GridViewColumnEx), new PropertyMetadata(false));
        #endregion

        #region CanHide (DP SHORT) [false]
        public bool CanHide { get { return (bool)GetValue(CanHideProperty); } set { SetValue(CanHideProperty, value); } }
        public static readonly DependencyProperty CanHideProperty = DependencyProperty.Register(
            "CanHide", typeof(bool), typeof(GridViewColumnEx), new PropertyMetadata(false));
        #endregion

        #region CanMove (DP SHORT) [true]
        public bool CanMove { get { return (bool)GetValue(CanMoveProperty); } set { SetValue(CanMoveProperty, value); } }
        public static readonly DependencyProperty CanMoveProperty = DependencyProperty.Register(
            "CanMove", typeof(bool), typeof(GridViewColumnEx), new PropertyMetadata(true));
        #endregion

        #region CanResize (DP SHORT) [true]
        public bool CanResize { get { return (bool)GetValue(CanResizeProperty); } set { SetValue(CanResizeProperty, value); } }
        public static readonly DependencyProperty CanResizeProperty = DependencyProperty.Register(
            "CanResize", typeof(bool), typeof(GridViewColumnEx), new PropertyMetadata(true));
        #endregion

        #endregion

        #region DP ToolTip

        #region HideToolTip (DP SHORT)
        public string HideToolTip { get { return (string)GetValue(HideToolTipProperty); } set { SetValue(HideToolTipProperty, value); } }
        public static readonly DependencyProperty HideToolTipProperty = DependencyProperty.Register("HideToolTip",
            typeof(string), typeof(GridViewColumnEx),
            new PropertyMetadata("Hide column"));
        #endregion

        #region FilterToolTip (DP SHORT)
        public string FilterToolTip { get { return (string)GetValue(FilterToolTipProperty); } set { SetValue(FilterToolTipProperty, value); } }
        public static readonly DependencyProperty FilterToolTipProperty = DependencyProperty.Register("FilterToolTip",
            typeof(string), typeof(GridViewColumnEx),
            new PropertyMetadata("Filter column"));
        #endregion

        #endregion

        #region DP other

        #region HorizontalContentAlignment (DP SHORT)
        public HorizontalAlignment HorizontalContentAlignment { get { return (HorizontalAlignment)GetValue(HorizontalContentAlignmentProperty); } set { SetValue(HorizontalContentAlignmentProperty, value); } }
        public static readonly DependencyProperty HorizontalContentAlignmentProperty = DependencyProperty.Register("HorizontalContentAlignment", typeof(HorizontalAlignment), typeof(GridViewColumnEx), new PropertyMetadata(HorizontalAlignment.Left));
        #endregion

        #region DisableMoveColor (DP SHORT)
        public Brush DisableMoveColor { get { return (Brush)GetValue(DisableMoveColorProperty); } set { SetValue(DisableMoveColorProperty, value); } }
        public static readonly DependencyProperty DisableMoveColorProperty = DependencyProperty.Register("DisableMoveColor",
            typeof(Brush), typeof(GridViewColumnEx),
            new PropertyMetadata(new BrushConverter().ConvertFromString("#10000000")));
        #endregion

        #region FilterIndicatorColor (DP SHORT)
        public Brush FilterIndicatorColor { get { return (Brush)GetValue(FilterIndicatorColorProperty); } set { SetValue(FilterIndicatorColorProperty, value); } }
        public static readonly DependencyProperty FilterIndicatorColorProperty = DependencyProperty.Register("FilterIndicatorColor",
            typeof(Brush), typeof(GridViewColumnEx),
            new PropertyMetadata(Brushes.Green));
        #endregion

        #region MinColumnWidth (DP SHORT)
        public double MinColumnWidth { get { return (double)GetValue(MinColumnWidthProperty); } set { SetValue(MinColumnWidthProperty, value); } }
        public static readonly DependencyProperty MinColumnWidthProperty = DependencyProperty.Register("MinColumnWidth", typeof(double), typeof(GridViewColumnEx), new PropertyMetadata(50d));
        #endregion

        #endregion

        #region Profil (DP SHORT)
        public string Profil { get { return (string)GetValue(ProfilProperty); } set { SetValue(ProfilProperty, value); } }
        public static readonly DependencyProperty ProfilProperty = DependencyProperty.Register("Profil", typeof(string), typeof(GridViewColumnEx), new PropertyMetadata(null));
        #endregion

        #region ColumnBindingName (DP SHORT)
        public string ColumnBindingName { get { return (string)GetValue(ColumnBindingNameProperty); } set { SetValue(ColumnBindingNameProperty, value); } }
        public static readonly DependencyProperty ColumnBindingNameProperty = DependencyProperty.Register("ColumnBindingName",
            typeof(string), typeof(GridViewColumnEx),
            new PropertyMetadata(null, (ss, ee) =>
            {
                var gvc = ss as GridViewColumnEx;
                if (gvc.DisplayMemberBinding == null)
                {
                    gvc.DisplayMemberBinding = new Binding((string)ee.NewValue);
                }
            }));
        #endregion

        #region FilterEditor (DP SHORT)
        public FilterEditor FilterEditor { get { return (FilterEditor)GetValue(FilterEditorProperty); } set { SetValue(FilterEditorProperty, value); } }
        public static readonly DependencyProperty FilterEditorProperty = DependencyProperty.Register("FilterEditor",
            typeof(FilterEditor), typeof(GridViewColumnEx), new PropertyMetadata(FilterEditor.SingleType));
        #endregion
    }

    public class GridViewColumnHeaderEx : GridViewColumnHeader, INotifyPropertyChanged
    {
        #region INotifyPropertyChanged
        public event PropertyChangedEventHandler PropertyChanged;

        public void OnPropertyChanged([CallerMemberName] string propertyName = null)
        {
            this.OnPropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }
        protected virtual void OnPropertyChanged(object sender, PropertyChangedEventArgs args)
        {
            PropertyChanged?.Invoke(sender, args);
        }
        public void Raise(string propertyName)
        {
            OnPropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }
        #endregion

        #region Consctructors
        static GridViewColumnHeaderEx()
        {
            DefaultStyleKeyProperty.OverrideMetadata(typeof(GridViewColumnHeaderEx),
                new FrameworkPropertyMetadata(typeof(GridViewColumnHeaderEx)));
        }

        protected override void OnPropertyChanged(DependencyPropertyChangedEventArgs e)
        {
            base.OnPropertyChanged(e);
            if (e.Property.Name.Equals("ActualWidth"))
            {
                if (!IsHide && !IsHideProfil)
                {
                    double? pa = ColumnsHelper.GetMinColumnsWidth(ItemsControl);
                    double m = pa == null ? MinColumnWidth : (double)pa;
                    if (ActualWidth < m && !double.IsNaN(Column.Width))
                        Column.Width = m;
                }
            }
        }

        public GridViewColumnHeaderEx() { }
        #endregion

        #region Apply Template
        protected Popup popup = null;

        public override void OnApplyTemplate()
        {
            base.OnApplyTemplate();

            #region get Controls & lambda for focus
            popup = this.GetTemplateChild("PART_Popup") as Popup;
            if (popup != null)
            {
                Border uie = popup.Child as Border;
                uie.IsVisibleChanged += delegate
                {
                    var cc = GetChildOfType<ContentPresenter>(uie);
                    this.Dispatcher.BeginInvoke(System.Windows.Threading.DispatcherPriority.ApplicationIdle,
                        (Action)(() =>
                        {
                            var tb = GetFirstChildFocusable(cc);
                            tb?.Focus();
                        }));
                };
                popup.Opened += delegate { };
            }
            #endregion

            #region necessary for creation ColumnManager/FilterManager
            var cm = ColumnManager;
            var fm = FilterManager;
            #endregion

            #region Set editor
            // Find editor from type only canfilter and columName != null
            // check if columname exist in itemsource object otherwize change 
            if (CanFilter && ColumnBindingName != null)
            {
                var itemsControl = GetVisualParent<ItemsControl>(this);
                EventHandler ev = (s, e) =>
                {
                    FilterEditor fe = this.FilterEditor;
                    Editor = fm.FromName(this.ColumnBindingName, fe);
                    if (Editor == null)
                        CanFilter = false;
                    else
                    {
                        Editor.FilterChanged += delegate { IsFiltered = Editor.IsFilter; };
                        fm.Add(Editor);
                    }
                };
                // set itemsource changed
                var dpd = DependencyPropertyDescriptor.FromProperty(ItemsControl.ItemsSourceProperty, itemsControl.GetType());
                if (dpd != null)
                {
                    dpd.AddValueChanged(itemsControl, ev);
                }
                // check itemsource is setted
                if (itemsControl.ItemsSource != null)
                {
                    ev.Invoke(this, null);
                }
            }
            #endregion
        }
        #endregion

        #region Context Menu
        protected override void OnMouseRightButtonDown(MouseButtonEventArgs e)
        {
            base.OnMouseRightButtonDown(e);
            SetContextMenu();
        }

        private void SetContextMenu()
        {
            var cm = new ContextMenu();

            // select profils
            if (Keyboard.IsKeyDown(Key.LeftCtrl) || Keyboard.IsKeyDown(Key.RightCtrl))
            {
                Profil[] ps = ColumnManager.Profils;
                if (ps.Length != 0)
                {
                    if(ColumnManager.SelectedProfil != null)
                    {
                        ICommand clear = new RelayCommand(
                        param => ColumnManager.ClearProfil(),
                        param => true);
                        var mi = new MenuItem
                        {
                            Header = ColumnsTextHelper.GetClearProfilText(ItemsControl),
                            IsCheckable = false,
                            IsEnabled = true,
                            Command = clear,                           
                        };
                        cm.Items.Add(mi);
                        cm.Items.Add(new Separator());
                    }
                    ICommand cmd = new RelayCommand(
                        param => ColumnManager.SetProfil(param as Profil),
                        param => !ColumnManager.SelectedProfil?.Value.Equals((param as Profil)?.Value) ?? true);

                    foreach (var item in ps)
                    {
                        #region menu item
                        var mi = new MenuItem
                        {
                            Header = item.Name,
                            IsCheckable = false,
                            IsChecked = ColumnManager.SelectedProfil?.Value.Equals(item.Value) ?? false,
                            IsEnabled = true,
                            Command = cmd,
                            CommandParameter = item,
                        };
                        cm.Items.Add(mi);
                        #endregion
                    }
                }
            }
            // show columns with profil
            else if (Keyboard.IsKeyDown(Key.LeftShift) || Keyboard.IsKeyDown(Key.RightShift))
            {
                Grid.SetIsSharedSizeScope(cm, true);
                foreach (var col in RowPresenter.Columns)
                {
                    #region get texts
                    string header = null;
                    string profils = null;
                    if (col.Header is string s) header = s;
                    else if (col.Header is GridViewColumnHeaderEx ch)
                    {
                        header = ch.Content.ToString().Trim();
                        profils = ch.StringProfils;
                    }
                    #endregion

                    #region grid create
                    Grid grid = new Grid();
                    ColumnDefinition c1 = new ColumnDefinition();
                    c1.Width = new GridLength(20, GridUnitType.Auto);
                    c1.SharedSizeGroup = "a";
                    ColumnDefinition c2 = new ColumnDefinition();
                    c2.Width = new GridLength(4, GridUnitType.Pixel);
                    ColumnDefinition c3 = new ColumnDefinition();
                    c3.Width = new GridLength(20, GridUnitType.Star);
                    grid.ColumnDefinitions.Add(c1);
                    grid.ColumnDefinitions.Add(c2);
                    grid.ColumnDefinitions.Add(c3);
                    TextBlock th = new TextBlock();
                    th.Text = header;
                    grid.Children.Add(th);
                    TextBlock tp = new TextBlock();
                    tp.Text = profils;
                    Grid.SetColumn(tp, 2);
                    grid.Children.Add(tp);
                    #endregion

                    #region menu item
                    var mi = new MenuItem
                    {
                        Header = grid,
                        IsEnabled = false,
                    };
                    cm.Items.Add(mi);
                    #endregion
                }
            }
            else
            {
                #region check is item isfiltered
                bool hasFilter = FilterManager.IsFiltered || IsFiltered;
                //hasFilter = rh.Columns.Where(h => h.Header is GridViewColumnHeaderEx)
                //                .Cast<GridViewColumnHeaderEx>()
                //                .Where(ch => ch.IsFiltered)
                //                .Count() != 0;           
                if (hasFilter)
                {
                    cm.Items.Add(new MenuItem
                    {
                        Header = ColumnsTextHelper.GetClearAllFilterText(ItemsControl),
                        Command = ClearFilterCommand,
                        CommandParameter = RowPresenter,
                        IsEnabled = hasFilter,
                    });
                    cm.Items.Add(new MenuItem
                    {
                        Header = ColumnsTextHelper.GetClearFilterText(ItemsControl),
                        Command = ClearFilterCommand,
                        CommandParameter = this,
                        IsEnabled = IsFiltered,
                    });
                }
                #endregion

                #region colum can hided
                List<GridViewColumn> canHided = new List<GridViewColumn>();
                foreach (GridViewColumn item in RowPresenter.Columns)
                {
                    if (item.Header is GridViewColumnHeaderEx h)
                    {
                        if (h.CanHide)
                        {
                            canHided.Add(item);
                        }
                    }
                }
                #endregion

                #region add separator if filter and element can hided
                if (hasFilter && canHided.Count() != 0)
                    cm.Items.Add(new Separator());
                #endregion

                #region adding current profile selected
                if (ColumnManager.SelectedProfil != null)
                {
                    var pt = ColumnsTextHelper.GetProfilText(ItemsControl);
                    cm.Items.Add(new MenuItem
                    {
                        Header = $"{pt} : {ColumnManager.SelectedProfil.Name}",
                        IsEnabled = false,
                    });
                    cm.Items.Add(new Separator());
                }
                #endregion

                #region add can hided
                foreach (var item in canHided)
                {
                    var h = item.Header as GridViewColumnHeaderEx;

                    var enabled = !h.HasProfil ||
                            ColumnManager.SelectedProfil == null ||
                            h.IsInProfil(ColumnManager.SelectedProfil);

                    var mi = new MenuItem
                    {
                        Header = h.Content.ToString(),
                        Command = ChangeHideCommand,
                        CommandParameter = item,
                        IsEnabled = enabled,
                        IsCheckable = !enabled,
                        IsChecked = enabled && !h.IsHide,
                    };
                    cm.Items.Add(mi);
                }
                #endregion
            }
            // show context if not empty
            if (cm.Items.Count != 0)
            {
                this.ContextMenu = cm;
            }
        }
#endregion

        #region Profil Utils
        public bool HasProfil => !string.IsNullOrEmpty(Profil);
        public bool IsInProfil(Profil profil) => IsInProfil(profil.Value);
        public bool IsInProfil(string checkName)
        {
            if (!HasProfil) return false;
            else return Profil.Split(',')
                .Where(x => x.Equals(checkName))
                .Count() != 0;
        }
        public string StringProfils
        {
            get
            {
                if (!HasProfil) return null;
                Profil[] ps = ColumnManager.Profils;
                var x = Profil.Split(',')
                            .Select(x => $"[{ps.FromValue(x)}] ")
                            .Aggregate((x, y) => x + y);
                return x;
            }
        }
        #endregion

        #region Width / IsHide / IsHideProfil

        #region PreviousWidth
        private double? _PreviousWidth = null;
        public double? PreviousWidth
        {
            get => _PreviousWidth;
            set
            {
                _PreviousWidth = value;
                //PropertyChanged?.Invoke(this, new PropertyChangedEventArgs("PreviousWidth"));
            }
        }
        #endregion

        #region IsHide
        private bool _IsHide = false;
        public bool IsHide
        {
            get => _IsHide;
            set
            {
                _IsHide = value;
                UpdateWidth();
                PropertyChanged?.Invoke(this, new PropertyChangedEventArgs("IsHide"));
            }
        }
        #endregion

        #region IsHideProfil
        private bool _IsHideProfil = false;
        public bool IsHideProfil
        {
            get => _IsHideProfil;
            set
            {
                _IsHideProfil = value;
                UpdateWidth();
                PropertyChanged?.Invoke(this, new PropertyChangedEventArgs("IsHideProfil"));
            }
        }
        #endregion

        private void UpdateWidth()
        {
            if (IsHide || IsHideProfil)
            {
                if (PreviousWidth == null)
                {
                    PreviousWidth = Column.ActualWidth;
                    Column.Width = 0;
                    ColumnManager.CheckSort(this);
                }
            }
            else
            {
                if (PreviousWidth != null)
                {
                    double x = (double)PreviousWidth;
                    PreviousWidth = null;
                    Column.Width = x;
                }
            }
        }

        #endregion
     
        #region TOOLS ColumnManager / FilterManager
        private GridViewColumn GetFromHeader(GridViewColumnHeaderEx header)
        {
            foreach (var item in RowPresenter.Columns)
            {
                if (item.Header == header) return item;
            }
            return null;
        }
        private List<GridViewColumn> GetListFromHeader =>
            RowPresenter.Columns.ToList();
        private GridViewHeaderRowPresenter RowPresenter =>
            GetVisualParent<GridViewHeaderRowPresenter>(this);
        private ItemsControl ItemsControl =>
            GetVisualParent<ItemsControl>(this);

        private ColumnsManager _ColumnManager = null;
        public ColumnsManager ColumnManager
        {
            get
            {
                if (_ColumnManager == null)
                {
                    var r = RowPresenter;
                    var ic = ItemsControl;
                    _ColumnManager = GetColumnManager(r);
                    if (_ColumnManager == null)
                    {
                        _ColumnManager = new ColumnsManager(ic, r);
                        SetColumnManager(r, _ColumnManager);
                    }
                }
                return _ColumnManager;
            }
        }

        private FilterManager _FilterManager = null;
        public FilterManager FilterManager
        {
            get
            {
                if (_FilterManager == null)
                {
                    var r = RowPresenter;
                    var ic = ItemsControl;
                    _FilterManager = GetFilterManager(r);
                    if (_FilterManager == null)
                    {
                        _FilterManager = new FilterManager(ic, r);
                        SetFilterManager(r, _FilterManager);
                    }
                }
                return _FilterManager;
            }
        }

        #region ColumnManager (DPS)
        public static ColumnsManager GetColumnManager(DependencyObject obj) => (ColumnsManager)obj.GetValue(ColumnManagerProperty);
        public static void SetColumnManager(DependencyObject obj, ColumnsManager value) => obj.SetValue(ColumnManagerProperty, value);
        private static readonly DependencyProperty ColumnManagerProperty =
            DependencyProperty.RegisterAttached("ColumnManager",
            typeof(ColumnsManager),
            typeof(GridViewColumnHeaderEx),
            new PropertyMetadata(null));
        #endregion

        #region FilterManager (DPS)             
        public static FilterManager GetFilterManager(DependencyObject obj) => (FilterManager)obj.GetValue(FilterManagerProperty);
        public static void SetFilterManager(DependencyObject obj, FilterManager value) => obj.SetValue(FilterManagerProperty, value);
        public static readonly DependencyProperty FilterManagerProperty =
            DependencyProperty.RegisterAttached("FilterManager",
            typeof(FilterManager),
            typeof(GridViewColumnHeaderEx),
            new PropertyMetadata(null));
        #endregion

        #endregion

        #region disable move header
        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs e)
        {
            base.OnMouseLeftButtonDown(e);
            if (!CanMove)
            {
                e.Handled = true;
            }
        }
        #endregion

        #region moving header protection (distinct move and click)
        Point clickPoint = new Point();
        bool onMoving = false;
        protected override void OnPreviewMouseLeftButtonUp(MouseButtonEventArgs e)
        {
            base.OnMouseLeftButtonUp(e);
            if (!onMoving && CanSort &&
                !string.IsNullOrEmpty(ColumnBindingName) &&
                e.OriginalSource is GridViewColumnHeaderEx)
            {
                //Debug.WriteLine($"Column sort name = {ColumnBindingName} composant = {e.OriginalSource}");
                ColumnManager.Sort(this.ColumnBindingName);
            }
            onMoving = false;
        }
        protected override void OnPreviewMouseLeftButtonDown(MouseButtonEventArgs e)
        {
            base.OnPreviewMouseLeftButtonDown(e);
            clickPoint = Mouse.GetPosition(this);
            //Debug.WriteLine($"initial point = {clickPoint}");
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);
            if (Mouse.LeftButton == MouseButtonState.Pressed)
            {
                var vec = Point.Subtract(clickPoint, Mouse.GetPosition(this));
                //Debug.WriteLine($"difference = {vec}");
                if (Math.Abs(vec.X) > 10 || Math.Abs(vec.Y) > 10)
                    onMoving = true;
            }
        }
        #endregion

        #region DP 

        #region CanSort (DP SHORT) [false]
        public bool CanSort { get { return (bool)GetValue(CanSortProperty); } set { SetValue(CanSortProperty, value); } }
        public static readonly DependencyProperty CanSortProperty = DependencyProperty.Register(
            "CanSort", typeof(bool), typeof(GridViewColumnHeaderEx), new PropertyMetadata(false));
        #endregion

        #region CanFilter (DP SHORT) [false]
        public bool CanFilter { get { return (bool)GetValue(CanFilterProperty); } set { SetValue(CanFilterProperty, value); } }
        public static readonly DependencyProperty CanFilterProperty = DependencyProperty.Register(
            "CanFilter", typeof(bool), typeof(GridViewColumnHeaderEx), new PropertyMetadata(false));
        #endregion

        #region CanHide (DP SHORT) [false]
        public bool CanHide { get { return (bool)GetValue(CanHideProperty); } set { SetValue(CanHideProperty, value); } }
        public static readonly DependencyProperty CanHideProperty = DependencyProperty.Register(
            "CanHide", typeof(bool), typeof(GridViewColumnHeaderEx), new PropertyMetadata(false));
        #endregion

        #region CanMove (DP SHORT) [true]
        public bool CanMove { get { return (bool)GetValue(CanMoveProperty); } set { SetValue(CanMoveProperty, value); } }
        public static readonly DependencyProperty CanMoveProperty = DependencyProperty.Register(
            "CanMove", typeof(bool), typeof(GridViewColumnHeaderEx), new PropertyMetadata(true));
        #endregion

        #region CanResize (DP SHORT) [true]
        public bool CanResize { get { return (bool)GetValue(CanResizeProperty); } set { SetValue(CanResizeProperty, value); } }
        public static readonly DependencyProperty CanResizeProperty = DependencyProperty.Register(
            "CanResize", typeof(bool), typeof(GridViewColumnHeaderEx), new PropertyMetadata(true));
        #endregion

        #region HideToolTip (DP SHORT)
        public string HideToolTip { get { return (string)GetValue(HideToolTipProperty); } set { SetValue(HideToolTipProperty, value); } }
        public static readonly DependencyProperty HideToolTipProperty = DependencyProperty.Register("HideToolTip",
            typeof(string), typeof(GridViewColumnHeaderEx),
            new PropertyMetadata("Masquer la colonne"));
        #endregion

        #region FilterToolTip (DP SHORT)
        public string FilterToolTip { get { return (string)GetValue(FilterToolTipProperty); } set { SetValue(FilterToolTipProperty, value); } }
        public static readonly DependencyProperty FilterToolTipProperty = DependencyProperty.Register("FilterToolTip",
            typeof(string), typeof(GridViewColumnHeaderEx),
            new PropertyMetadata("Filtrer"));
        #endregion
   
        #region Profil (DP SHORT)
        public string Profil { get { return (string)GetValue(ProfilProperty); } set { SetValue(ProfilProperty, value); } }
        public static readonly DependencyProperty ProfilProperty = DependencyProperty.Register("Profil",
            typeof(string), typeof(GridViewColumnHeaderEx),
            new PropertyMetadata(null));
        #endregion

        #region DisableMoveColor (DP SHORT)
        public Brush DisableMoveColor { get { return (Brush)GetValue(DisableMoveColorProperty); } set { SetValue(DisableMoveColorProperty, value); } }
        public static readonly DependencyProperty DisableMoveColorProperty = DependencyProperty.Register("DisableMoveColor",
            typeof(Brush), typeof(GridViewColumnHeaderEx),
            new PropertyMetadata(new BrushConverter().ConvertFromString("#10000000")));
        #endregion

        #region FilterIndicatorColor (DP SHORT)
        public Brush FilterIndicatorColor { get { return (Brush)GetValue(FilterIndicatorColorProperty); } set { SetValue(FilterIndicatorColorProperty, value); } }
        public static readonly DependencyProperty FilterIndicatorColorProperty = DependencyProperty.Register("FilterIndicatorColor",
            typeof(Brush), typeof(GridViewColumnHeaderEx),
            new PropertyMetadata(Brushes.Green));
        #endregion

        #region ColumnBindingName (DP SHORT)
        public string ColumnBindingName { get { return (string)GetValue(ColumnBindingNameProperty); } set { SetValue(ColumnBindingNameProperty, value); } }
        public static readonly DependencyProperty ColumnBindingNameProperty = DependencyProperty.Register("ColumnBindingName",
            typeof(string), typeof(GridViewColumnHeaderEx),
            new PropertyMetadata(null, (s, e) =>
            {
                int vv = 0;
            }));
        #endregion

        #region SortDirection (DP SHORT)
        public ListSortDirectionExtended SortDirection { get { return (ListSortDirectionExtended)GetValue(SortDirectionProperty); } set { SetValue(SortDirectionProperty, value); } }
        public static readonly DependencyProperty SortDirectionProperty = DependencyProperty.Register("SortDirection",
            typeof(ListSortDirectionExtended), typeof(GridViewColumnHeaderEx),
            new PropertyMetadata(ListSortDirectionExtended.None));
        #endregion

        #region IsFiltered (DP SHORT)
        public bool IsFiltered { get { return (bool)GetValue(IsFilteredProperty); } set { SetValue(IsFilteredProperty, value); } }
        public static readonly DependencyProperty IsFilteredProperty = DependencyProperty.Register("IsFiltered",
            typeof(bool), typeof(GridViewColumnHeaderEx),
            new PropertyMetadata(false));
        #endregion

        #region FilterEditor (DP SHORT)
        public FilterEditor FilterEditor { get { return (FilterEditor)GetValue(FilterEditorProperty); } set { SetValue(FilterEditorProperty, value); } }
        public static readonly DependencyProperty FilterEditorProperty = DependencyProperty.Register("FilterEditor",
            typeof(FilterEditor), typeof(GridViewColumnHeaderEx),
            new PropertyMetadata(FilterEditor.SingleType));
        #endregion

        #region MinColumnWidth (DP SHORT)
        public double MinColumnWidth { get { return (double)GetValue(MinColumnWidthProperty); } set { SetValue(MinColumnWidthProperty, value); } }
        public static readonly DependencyProperty MinColumnWidthProperty = DependencyProperty.Register("MinColumnWidth", typeof(double), typeof(GridViewColumnHeaderEx), new PropertyMetadata(50d));
        #endregion

        #region Editor (DP SHORT)
        public BaseFilter Editor { get { return (BaseFilter)GetValue(EditorProperty); } set { SetValue(EditorProperty, value); } }
        public static readonly DependencyProperty EditorProperty = DependencyProperty.Register("Editor", typeof(BaseFilter), typeof(GridViewColumnHeaderEx), new PropertyMetadata(null));
        #endregion

        #endregion

        #region COMMAND

        #region ClearFilter (COMMAND)
        RelayCommand _ClearFilterCommand;
        public ICommand ClearFilterCommand
        {
            get
            {
                if (_ClearFilterCommand == null)
                {
                    _ClearFilterCommand = new RelayCommand(param =>
                    {
                        if (param is GridViewHeaderRowPresenter gvrp)
                        {
                            FilterManager.ClearAllFilter();
                        }
                        else if (param is GridViewColumnHeaderEx gvch)
                        {
                            gvch.Editor.Clear();
                        }
                    },
                        param => true);
                }
                return _ClearFilterCommand;
            }
        }
        #endregion

        #region ChangeHide (COMMAND)
        RelayCommand _ChangeHideCommand;
        public ICommand ChangeHideCommand
        {
            get
            {
                if (_ChangeHideCommand == null)
                {
                    _ChangeHideCommand = new RelayCommand(param =>
                    {
                        if (param is GridViewColumnHeaderEx hh)
                        {
                            param = GetFromHeader(hh);
                        }
                        if (param is GridViewColumn c)
                        {
                            if (c.Header is GridViewColumnHeaderEx h)
                            {
                                h.IsHide = !h.IsHide;
                            }
                        }
                    },
                        param => true);
                }
                return _ChangeHideCommand;
            }
        }
        #endregion

        #endregion

        #region STATIC
        private static T GetVisualParent<T>(DependencyObject child) where T : FrameworkElement
        {
            while ((child != null) && !(child is T))
            {
                child = VisualTreeHelper.GetParent(child);
            }
            return child as T;
        }
        private static T GetChildOfType<T>(DependencyObject depObj) where T : DependencyObject
        {
            if (depObj == null) return null;
            for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
            {
                var child = VisualTreeHelper.GetChild(depObj, i);

                var result = (child as T) ?? GetChildOfType<T>(child);
                if (result != null) return result;
            }
            return null;
        }
        private static UIElement GetFirstChildFocusable(UIElement depObj)
        {
            if (depObj == null) return null;
            for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
            {
                var child = VisualTreeHelper.GetChild(depObj, i) as UIElement;

                if (child?.Focusable ?? false) return child;

                var result = GetFirstChildFocusable(child);
                if (result != null) return result;
            }
            return null;
        }
        #endregion

        public Visible Visible => (IsHide ? Visible.Hide : Visible.Visible) 
                                    | (IsHideProfil ? Visible.HideProfil : Visible.Visible);
    }

    [Flags]
    public enum Visible
    {
        Visible = 0x0000,
        Hide = 0x0001,
        HideProfil = 0x0002,
    }

    public static class ProfilsArrayExtension
    {
        public static Profil FromValue(this Profil[] ps, string value) =>
               ps.Where(x => x.Value.Equals(value)).FirstOrDefault();
    }
}

