﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;
using System.Windows.Media;

namespace TestGridViewColumEx
{
    public class ColumnsManager : INotifyPropertyChanged
    {
        public event PropertyChangedEventHandler? PropertyChanged;

        #region private properties & constructor
        ItemsControl itemsControl = null;
        GridViewHeaderRowPresenter rowPresenter = null;

        public ColumnsManager(ItemsControl ic, GridViewHeaderRowPresenter rp)
        {
            itemsControl = ic;
            rowPresenter = rp;
        }

        private IEnumerable<object> ItemsSource => itemsControl?.ItemsSource as IEnumerable<object>;
        #endregion

        #region SORT function + utils
        public string SortName { get; private set; } = null;
        private ListSortDirectionExtended sortDirection = ListSortDirectionExtended.None;

        public bool IsSorted => sortDirection != ListSortDirectionExtended.None;

        #region ClearSort (SATIC COMMAND)
        static RelayCommand _ClearSortCommand;
        public static ICommand ClearSortCommand
        {
            get
            {
                if (_ClearSortCommand == null)
                {
                    _ClearSortCommand = new RelayCommand(param =>
                    {
                        if (param is DependencyObject dpo)
                        {
                            if (GetChildOfType<GridViewHeaderRowPresenter>(dpo) is GridViewHeaderRowPresenter rp)
                            {
                                ColumnsManager sm = GridViewColumnHeaderEx.GetColumnManager(rp);
                                sm.Sort(null);
                            }
                        }
                    },
                        param =>
                        {
                            if (param is DependencyObject dpo)
                            {
                                if (GetChildOfType<GridViewHeaderRowPresenter>(dpo) is GridViewHeaderRowPresenter rp)
                                {
                                    ColumnsManager sm = GridViewColumnHeaderEx.GetColumnManager(rp);
                                    return sm?.IsSorted ?? false;
                                }
                            }
                            return false;
                        });
                }
                return _ClearSortCommand;
            }
        }
        #endregion

        /// <summary>
        /// for sorting
        /// - if columName = null clear sorting
        /// - if columName = previousColumnName and clear = true, clear sort
        /// - if columName != previousColumnName and clear = true, not clear sort
        /// otherwise sorting
        /// - if previousColumnName = null -> descending
        /// - if columName = previousColumnName and descending -> ascending
        /// - if columName = previousColumnName and ascending -> none -> clear sort
        /// </summary>
        /// <param name="columName"></param>
        /// <param name="clear"></param>
        public void Sort(string columName, bool clear = false)
        {
            if (clear)
            {
                foreach (var item in rowPresenter.Columns)
                {
                    if (item.Header is GridViewColumnHeaderEx h)
                    {
                        if (columName?.Equals(h.ColumnBindingName) ?? false)
                        {
                            if (h.SortDirection != ListSortDirectionExtended.None)
                            {
                                h.SortDirection = ListSortDirectionExtended.None;
                                SortName = null;
                                sortDirection = h.SortDirection;
                                ICollectionView view = CollectionViewSource.GetDefaultView(ItemsSource);
                                view.SortDescriptions.Clear();
                            }
                        }
                    }
                }
            }
            else
            {
                foreach (var item in rowPresenter.Columns)
                {
                    if (item.Header is GridViewColumnHeaderEx h)
                    {
                        if (columName?.Equals(h.ColumnBindingName) ?? false)
                        {
                            if (!columName.Equals(SortName))
                            {
                                h.SortDirection = ListSortDirectionExtended.Descending;
                                SortName = columName;
                                sortDirection = h.SortDirection;
                            }
                            else
                            {
                                if (h.SortDirection == ListSortDirectionExtended.Ascending)
                                {
                                    h.SortDirection = ListSortDirectionExtended.None;
                                    SortName = null;
                                    sortDirection = h.SortDirection;
                                }
                                else
                                {
                                    h.SortDirection = ListSortDirectionExtended.Ascending;
                                    sortDirection = h.SortDirection;
                                }
                            }
                            // set sort 
                            ICollectionView view = CollectionViewSource.GetDefaultView(ItemsSource);
                            view.SortDescriptions.Clear();
                            if (h.SortDirection != ListSortDirectionExtended.None)
                            {
                                view.SortDescriptions.Add(new SortDescription(SortName,
                                    h.SortDirection == ListSortDirectionExtended.Ascending ?
                                    System.ComponentModel.ListSortDirection.Ascending :
                                    System.ComponentModel.ListSortDirection.Descending));
                            }
                        }
                        else
                        {
                            h.SortDirection = ListSortDirectionExtended.None;
                        }
                    }
                }
                if (columName == null)
                {
                    SortName = null;
                    sortDirection = ListSortDirectionExtended.None;
                    ICollectionView view = CollectionViewSource.GetDefaultView(ItemsSource);
                    view.SortDescriptions.Clear();
                }
            }
        }

        public void CheckSort(GridViewColumnHeaderEx gvch)
        {
            if ((gvch.IsHide || gvch.IsHideProfil) && gvch.ColumnBindingName.Equals(SortName))
                Sort(null);
        }
        #endregion

        #region Profils + utils
        
        private Profil[] _Profils = null;
        public Profil[] Profils
        {
            get
            {
                if (_Profils == null)
                {
                    IValueConverter converter = ProfilHelper.GetConverter(itemsControl);
                    List<Profil> ps = new List<Profil>();
                    //var lp = rowPresenter.Columns
                    var lp = rowPresenter.Columns
                        .Where(x => x.Header is GridViewColumnHeaderEx)
                        .Select(x => x.Header)
                        .Cast<GridViewColumnHeaderEx>()
                        .Where(x => !string.IsNullOrEmpty(x.Profil))
                        .SelectMany(x => x.Profil.Split(','))
                        .Distinct()
                        .Select(x => new Profil(x, (string)converter?.Convert(x, null, null, null)))
                        .OrderBy(x => x.Name)
                        .ToArray();

                    _Profils = lp.Length == 0 ? null : lp;
                }
                return _Profils;
            }
        }

        public static Profil[] GetProfils(DependencyObject dpo)
        {
            if (GetChildOfType<GridViewHeaderRowPresenter>(dpo) is GridViewHeaderRowPresenter rp)
            {
                ColumnsManager sm = GridViewColumnHeaderEx.GetColumnManager(rp);
                return sm.Profils;
            }
            return null;
        }
        public static void SetProfil(DependencyObject dpo, Profil profil)
        {
            if (GetChildOfType<GridViewHeaderRowPresenter>(dpo) is GridViewHeaderRowPresenter rp)
            {
                ColumnsManager sm = GridViewColumnHeaderEx.GetColumnManager(rp);
                sm.SetProfil(profil);
            }
        }
        public static void ClearProfil(DependencyObject dpo)
        {
            if (GetChildOfType<GridViewHeaderRowPresenter>(dpo) is GridViewHeaderRowPresenter rp)
            {
                ColumnsManager sm = GridViewColumnHeaderEx.GetColumnManager(rp);
                sm.ClearProfil();
            }
        }
        public static ColumnsManager Get(DependencyObject dpo)
        {
            if (GetChildOfType<GridViewHeaderRowPresenter>(dpo) is GridViewHeaderRowPresenter rp)
            {
                return GridViewColumnHeaderEx.GetColumnManager(rp);              
            }
            return null;
        }

        public void SetProfil(Profil profil) => SelectedProfil = profil;
        public void ClearProfil() => SelectedProfil = null;

        #region SelectedProfil
        private Profil _SelectedProfil = null;
        public Profil SelectedProfil
        {
            get => _SelectedProfil;
            set
            {              
                _SelectedProfil = value;
                // restore colum hided by profil
                if(value == null)
                {
                    rowPresenter.Columns
                        .Where(x => x.Header is GridViewColumnHeaderEx)
                        .Select(x => x.Header)
                        .Cast<GridViewColumnHeaderEx>()
                        .Where(x => x.IsHideProfil)
                        .ToList()
                        .ForEach(x => x.IsHideProfil = false);
                }
                // hide column has profil and not the selected profil
                else
                {
                    var lp = rowPresenter.Columns
                        .Where(x => x.Header is GridViewColumnHeaderEx)
                        .Select(x => x.Header)
                        .Cast<GridViewColumnHeaderEx>()
                        .Where(x => x.HasProfil);
                    foreach (var item in lp)
                    {
                        item.IsHideProfil = !item.IsInProfil(value.Value);
                        // disable sort 
                        if (item.IsHideProfil && item.ColumnBindingName.Equals(SortName))
                            Sort(null);
                    }                          
                }
                PropertyChanged?.Invoke(this, new PropertyChangedEventArgs("SelectedProfil"));
            }
        }
        #endregion

        #endregion

        #region Other function

        #region ResetColumnsWidth (STATIC COMMAND)
        static RelayCommand _ResetColumnsWidthCommand = null;
        public static ICommand ResetColumnsWidthCommand
        {
            get
            {
                if (_ResetColumnsWidthCommand == null)
                {
                    _ResetColumnsWidthCommand = new RelayCommand(param =>
                        {
                            if (param is DependencyObject dpo)
                            {
                                if (GetChildOfType<GridViewHeaderRowPresenter>(dpo) is GridViewHeaderRowPresenter rp)
                                {
                                    ColumnsManager sm = GridViewColumnHeaderEx.GetColumnManager(rp);
                                    sm.ResetColumsWidth();
                                }
                            }
                        },
                        param => true);
                }
                return _ResetColumnsWidthCommand;
            }
        }
        #endregion

        public void ResetColumsWidth()
        {
            foreach (var item in rowPresenter.Columns)
            {
                bool canDo = true;
                if (item.Header is GridViewColumnHeaderEx gvch)
                {
                    canDo = !gvch.IsHide && !gvch.IsHideProfil;
                }
                if(canDo) item.Width = Double.NaN;
            }
        }

        #endregion

        #region STATIC
        private static T GetChildOfType<T>(DependencyObject depObj) where T : DependencyObject
        {
            if (depObj == null) return null;
            for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
            {
                var child = VisualTreeHelper.GetChild(depObj, i);

                var result = (child as T) ?? GetChildOfType<T>(child);
                if (result != null) return result;
            }
            return null;
        }
        #endregion
    }

    public enum ListSortDirectionExtended
    {
        // no sort
        None = -1,
        // Ascending sort
        Ascending = 0,
        // Descending sort
        Descending = 1
    }

    public class Profil
    {
        public string Value { get; private set; } = null;
        private string _Name = null;
        public string Name
        {
            get => _Name ?? Value;
            private set => _Name = value;
        }

        public Profil(string value, string name = null)
        {
            Value = value;
            Name = name;
        }

        public override string ToString() => Name;     
    }
}
