﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;
using System.Windows.Media;

namespace TestGridViewColumEx
{
    [Flags]
    public enum FilterEditor
    {
        // sigle value for type int (for int, double, float) string (for string) 
        // one date, one hour (DEFAULT)
        // if set and value is enum -> change to Enum
        SingleType = 0x0000_0001,
        // two value date, time, numeric from to
        Two = 0x0000_0002,
        // for enum values (set automaticly if detect type enum)
        Enum = 0x0000_0004,
        // for show values as enum
        Values = 0x0000_0008,
        TypeGroup = 0x0000_000f,
        // for enum value with select all, select none
        All = 0x0000_0010,
        None = 0x0000_0020,
        AllNone = 0x0000_0030,
        SelectGroup = 0x0000_00f0,
        // for datetime
        OnlyDate = 0x0000_0000, // default
        AllDateTime = 0x0000_0100,
        DateTimeGroup = 0x0000_0f00,
        // for numeric 
        MatchExact = 0x0000_1000,
        // for datetime (having a specific editor)
        Specific = 0x0000_2000,
        // date format
        DMY = 0x0000_0000, // default
        MDY = 0x0001_0000,
        YMD = 0x0002_0000,
        YDM = 0x0004_0000,
        DateFormatGroup = 0x000f_0000,
    }

    public class FilterManager : List<BaseFilter>
    {
        #region constructor
        ItemsControl itemsControl = null;
        GridViewHeaderRowPresenter rowPresenter = null;
        ICollectionView view = null;
        public FilterManager(ItemsControl ic, GridViewHeaderRowPresenter rp)
        {
            itemsControl = ic;
            rowPresenter = rp;

            EventHandler ev = (s, e) =>
            {
                view = CollectionViewSource.GetDefaultView(ItemsSource);
                view.Filter = x =>
                {
                    var f = this.Where(a => a.IsFilter);
                    // no one is filtered
                    if (f.Count() == 0) return true;
                    // check all filter
                    foreach (var item in f)
                    {
                        if (!item.Check(x)) return false;
                    }
                    return true;
                };
            };

            var dpd = DependencyPropertyDescriptor.FromProperty(ItemsControl.ItemsSourceProperty, itemsControl.GetType());
            if (dpd != null)
            {
                dpd.AddValueChanged(itemsControl, ev);
            }
            if (ItemsSource != null) ev.Invoke(this, null);
        }

        private IEnumerable<object> ItemsSource => itemsControl?.ItemsSource as IEnumerable<object>;
        #endregion

        #region Add / Clear
        public new void Add(BaseFilter filter)
        {
            base.Add(filter);
            filter.FilterChanged += delegate { view.Refresh(); };
        }
        /// <summary>
        /// clear dictionnay except locked filter
        /// </summary>
        public new void Clear() => this.ForEach(x => { if (!x.Locked) x.Clear(); });
        #endregion

        #region STATIC COMMAND & inside command

        #region ClearAllFilter (COMMAND)
        private static RelayCommand _ClearAllFilterCommand;
        public static ICommand ClearAllFilterCommand
        {
            get
            {
                if (_ClearAllFilterCommand == null)
                {
                    _ClearAllFilterCommand = new RelayCommand(param =>
                    {
                        if (param is DependencyObject dpo)
                        {
                            if (GetChildOfType<GridViewHeaderRowPresenter>(dpo) is GridViewHeaderRowPresenter rp)
                            {
                                FilterManager fm = GridViewColumnHeaderEx.GetFilterManager(rp);
                                fm.ClearAllFilter();
                            }
                        }
                    },
                    param =>
                    {
                        if (param is DependencyObject dpo)
                        {
                            if (GetChildOfType<GridViewHeaderRowPresenter>(dpo) is GridViewHeaderRowPresenter rp)
                            {
                                FilterManager fm = GridViewColumnHeaderEx.GetFilterManager(rp);
                                return fm?.IsFiltered ?? false;
                            }
                        }
                        return false;
                    });
                }
                return _ClearAllFilterCommand;
            }
        }
        #endregion

        public void ClearAllFilter()
        {
            this.Where(x => x.IsFilter).ToList().ForEach(x => x.Clear());
        }      
        #endregion

        #region Get Editor functions
        public BaseFilter FromType(string column, Type type, FilterEditor fi = FilterEditor.SingleType)
        {
            var tc = Type.GetTypeCode(type);
            bool ok = false;
            // boolean 
            if (tc == TypeCode.Boolean)
                return new BoolFilter(column);
            // other
            if ((fi & FilterEditor.SingleType) != 0)
            {
                if (type.IsEnum) { }
                else if (tc == TypeCode.String)
                    return new StringFilter(column);
                else if (tc >= TypeCode.Int16 && tc <= TypeCode.Single)
                    return new IntegerFilter(column, (fi & FilterEditor.MatchExact) != 0);               
                else if (tc == TypeCode.DateTime)
                {
                    if ((fi & FilterEditor.Specific) != 0 || (fi & FilterEditor.MatchExact) != 0)
                        return new SpecificDateFilter(column);
                    else
                        return new StringDateFilter(column, (fi & FilterEditor.DateFormatGroup));
                }
            }
            else if ((fi & FilterEditor.Enum) != 0)
            {
                if (!type.IsEnum) return null;
            }
            else if ((fi & FilterEditor.Two) != 0)
            {
                if (tc >= TypeCode.Int16 && tc <= TypeCode.Single)
                    return new TwoIntegerFilter(column);
                else if (tc == TypeCode.DateTime)
                    return new TwoDateFilter(column);               
            }
            else if ((fi & FilterEditor.Values) != 0)
            {
                var enumerator = ItemsSource?.GetEnumerator();
                enumerator?.MoveNext();
                PropertyInfo? pi = enumerator?.Current?.GetType().GetProperty(column);
                if (tc == TypeCode.DateTime)
                {
                    Array a = null;
                    if ((fi & FilterEditor.AllDateTime) != 0)
                    {

                    }
                    else
                    {
                        a = ItemsSource.Select(x => pi.GetValue(x))
                                .Cast<DateTime>()
                                .Select(x => new DateTime(x.Year, x.Month, x.Day, 0, 0, 0))
                                .Distinct()
                                .OrderBy(x => x)
                                .ToArray();
                    }
                    return new DateEnumFilter(column, a, fi & FilterEditor.DateFormatGroup,
                        ColumnsTextHelper.GetEnumAllText(itemsControl),
                        ColumnsTextHelper.GetEnumNoneText(itemsControl),
                        (fi & FilterEditor.All) != 0,
                        (fi & FilterEditor.None) != 0);
                }
                else
                {
                    // for other 
                    var a = ItemsSource.Select(x => pi.GetValue(x))
                                .Distinct()
                                .OrderBy(x => x)
                                .ToArray();
                    if (tc == TypeCode.String)
                        return new StringEnumFilter(column, a,
                            ColumnsTextHelper.GetEnumAllText(itemsControl),
                            ColumnsTextHelper.GetEnumNoneText(itemsControl),
                            (fi & FilterEditor.All) != 0,
                            (fi & FilterEditor.None) != 0);
                    else if (tc >= TypeCode.Int16 && tc <= TypeCode.Single)
                        return new IntEnumFilter(column, a,
                            ColumnsTextHelper.GetEnumAllText(itemsControl),
                            ColumnsTextHelper.GetEnumNoneText(itemsControl),
                            (fi & FilterEditor.All) != 0,
                            (fi & FilterEditor.None) != 0);
                }
            }
            // enum case
            if (type.IsEnum)
            {
                return new EnumEnumFilter(column, type,
                    ColumnsTextHelper.GetEnumAllText(itemsControl),
                    ColumnsTextHelper.GetEnumNoneText(itemsControl),
                    (fi & FilterEditor.All) != 0,
                    (fi & FilterEditor.None) != 0);
            }
            return null;
        }
        public BaseFilter FromName(string column, FilterEditor fi = FilterEditor.SingleType)
        {
            var enumerator = ItemsSource?.GetEnumerator();
            enumerator?.MoveNext();
            PropertyInfo? pi = enumerator?.Current?.GetType().GetProperty(column);
            if (pi != null) return FromType(column, pi.PropertyType, fi);
            else return null;
        }

        #endregion

        public bool IsFiltered => this.Where(x => x.IsFilter).Count() != 0;

        #region STATIC
        private static T GetChildOfType<T>(DependencyObject depObj) where T : DependencyObject
        {
            if (depObj == null) return null;
            for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
            {
                var child = VisualTreeHelper.GetChild(depObj, i);

                var result = (child as T) ?? GetChildOfType<T>(child);
                if (result != null) return result;
            }
            return null;
        }
        #endregion
    }
}
