﻿#define SHOW

using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;

namespace TestGridViewColumEx
{
    public class ColumnsHelper
    {
        #region API
        [DllImport("kernel32")]
        private static extern long WritePrivateProfileString(string section,
          string key, string val, string filePath);
        [DllImport("kernel32")]
        private static extern int GetPrivateProfileString(string section,
                 string key, string def, StringBuilder retVal,
            int size, string filePath);

        private static void IniWrite(string path, string section, string key, string value)
        {
            WritePrivateProfileString(section, key, value, path);
        }
        private static string IniRead(string path, string section, string key)
        {
            StringBuilder temp = new StringBuilder(255);
            int i = GetPrivateProfileString(section, key, "", temp, 255, path);
            return temp.ToString();
        }
        #endregion

        #region const & record
        private const string KEY_COL = "cols";          // sort position
        private const string KEY_SIZ = "sizes";         // size of columns
        private const string KEY_PRE = "previous";      // previous sizes
        // col visible 
        // 0 - visible
        // 1 - hide
        // 2 - hideprofil
        // 3 - hide & hideprofil
        private const string KEY_VIS = "visible";
        // selected profil
        private const string KEY_PRO = "profil";

        private record struct Parameter(bool UseIni,
                     string Compagny,
                     string Product,
                     string FileIni,
                     RegistryKey? RegistryKey);
        #endregion

        #region static parameter (UseIni dfefault  = true)
        // set in app.cs or before first InitializeComponent();
        public static bool UseIni { get; set; } = true;
        #endregion

        #region SaveColumnPositionID              
        public static string GetSaveColumnPositionID(DependencyObject obj) => (string)obj.GetValue(SaveColumnPositionIDProperty);
        public static void SetSaveColumnPositionID(DependencyObject obj, string value) => obj.SetValue(SaveColumnPositionIDProperty, value);
        public static readonly DependencyProperty SaveColumnPositionIDProperty =
            DependencyProperty.RegisterAttached("SaveColumnPositionID",
                typeof(string),
                typeof(ColumnsHelper),
                new PropertyMetadata(null
                , (ss, ee) =>
                {
                    if (ss is ListView lv)
                    {
                        #region informations compagny & product
                        var param = new Parameter(UseIni, "", "", "", null);

                        var vi = FileVersionInfo.GetVersionInfo(Assembly.GetEntryAssembly().Location);
                        param.Compagny = vi.CompanyName ?? "[Common]";
                        param.Product = vi.ProductName;
                        #endregion

                        #region For IniFile
                        if (UseIni)
                        {
                            // localisation %APPDATA%\[compagny]\[product].ini
                            // [ID]
                            // cols = 3 | 2 | 1 | 0 |
                            // sizes = NaN | 64.4 | 208.17660377358487 | NaN |
                            var dirIni = System.IO.Path.Combine(
                                Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData),
                                $"{param.Compagny}");
                            if (!System.IO.Directory.Exists(dirIni))
                            {
                                System.IO.Directory.CreateDirectory(dirIni);
                            }
                            var fileNameIni = $"{param.Product}.ini";
                            param.FileIni = System.IO.Path.Combine(dirIni, fileNameIni);
#if SHOW
                            Debug.WriteLine($"Chemin = {param.FileIni}");
#endif
                        }
                        #endregion
                        #region use registry
                        else
                        {
                            // localisation Computer\HKEY_CURRENT_USER\SOFTWARE\[compagny]\[product]\[Id]
                            // key cols
                            // key sizes
                            var rk = Registry.CurrentUser.CreateSubKey(@$"SOFTWARE\{param.Compagny}");
                            param.RegistryKey = rk.CreateSubKey($"{param.Product}");
                        }
                        #endregion

                        #region Loaded
                        bool firstUpdated = true;                       
                        lv.Loaded += delegate
                        {
                            if (firstUpdated)
                            {
                                firstUpdated = false;
                                #region Get Column & create list
                                if (FindChild<GridViewHeaderRowPresenter>(lv) is GridViewHeaderRowPresenter gvhrp)
                                {
                                    #region set ColumID
                                    int i = 0;
                                    foreach (var c in gvhrp.Columns)
                                    {
                                        SetColumnID(c, i++);
                                    }
                                    #endregion

                                    string cols = "", sizes = "", previous = "", visible = "", profil = "";
                                    #region IniFile
                                    if (UseIni)
                                    {
                                        if (System.IO.File.Exists(param.FileIni))
                                        {
                                            (cols, sizes, previous, visible, profil) = ReadIni(param, (string)ee.NewValue);
                                        }
                                    }
                                    #endregion
                                    #region Use registry
                                    else
                                    {
                                        (cols, sizes, previous, visible, profil) = ReadRegistry(param, (string)ee.NewValue);
                                    }
                                    #endregion

                                    SetColumnsValues(gvhrp.Columns, cols, sizes, previous, visible, profil);

                                    #region Connect for change
                                    var pd = DependencyPropertyDescriptor.FromProperty(GridViewColumn.WidthProperty, typeof(GridViewColumn));
                                    foreach (var c in gvhrp.Columns)
                                    {
                                        pd.AddValueChanged(c, delegate
                                        {
                                            UpdateDatas(param, (string)ee.NewValue, gvhrp.Columns);
                                        });
                                    }
                                    // notification if change
                                    gvhrp.Columns.CollectionChanged += delegate
                                    {
                                        UpdateDatas(param, (string)ee.NewValue, gvhrp.Columns);
                                    };
                                    #endregion
                                };
                                #endregion  

                                #region - Local functions -----------------------------------------

                                void UpdateDatas(Parameter param, string id, GridViewColumnCollection collection)
                                {
                                    if (param.UseIni) { WriteIni(param, id, collection); }
                                    else { WriteRegistry(param, id, collection); }
                                }

                                #region For Ini file
                                (string col, string size, string previous, string visible, string profil)
                                    ReadIni(Parameter param, string id)
                                {
                                    string cols = IniRead(param.FileIni, id, KEY_COL);
                                    string sizes = IniRead(param.FileIni, id, KEY_SIZ);
                                    string previous = IniRead(param.FileIni, id, KEY_PRE);
                                    string visible = IniRead(param.FileIni, id, KEY_VIS);
                                    string profil = IniRead(param.FileIni, id, KEY_PRO);
#if SHOW
                                    Debug.WriteLine("- READ INI --------------------------------------");
                                    Debug.WriteLine($"colonnes = {cols}");
                                    Debug.WriteLine($"size = {sizes}");
                                    Debug.WriteLine($"previous = {previous}");
                                    Debug.WriteLine($"visible = {visible}");
                                    Debug.WriteLine($"profil = {profil}");
#endif
                                    return (cols, sizes, previous, visible, profil);
                                }
                                void WriteIni(Parameter param, string id, GridViewColumnCollection collection)
                                {
                                    var (cols, sizes, previous, visible, profil) = GetColumnsValues(collection);
                                    IniWrite(param.FileIni, id, KEY_COL, cols);
                                    IniWrite(param.FileIni, id, KEY_SIZ, sizes);
                                    IniWrite(param.FileIni, id, KEY_PRE, previous);
                                    IniWrite(param.FileIni, id, KEY_VIS, visible);
                                    IniWrite(param.FileIni, id, KEY_PRO, profil);
                                }
                                #endregion

                                #region For registry
                                (string col, string size, string previous, string visible, string profil) ReadRegistry(Parameter param, string id)
                                {
                                    RegistryKey rk = param.RegistryKey.CreateSubKey(id);
                                    string cols = (string)rk.GetValue(KEY_COL);
                                    string sizes = (string)rk.GetValue(KEY_SIZ);
                                    string previous = (string)rk.GetValue(KEY_PRE);
                                    string visible = (string)rk.GetValue(KEY_VIS);
                                    string profil = (string)rk.GetValue(KEY_PRO);
#if SHOW
                                    Debug.WriteLine("- READ REGISTRY --------------------------------------");
                                    Debug.WriteLine($"colonnes = {cols}");
                                    Debug.WriteLine($"size = {sizes}");
                                    Debug.WriteLine($"previous = {previous}");
                                    Debug.WriteLine($"visible = {visible}");
                                    Debug.WriteLine($"profil = {profil}");
#endif
                                    return (cols, sizes, previous, visible, profil);
                                }
                                void WriteRegistry(Parameter param, string id, GridViewColumnCollection collection)
                                {
                                    var (cols, sizes, previous, visible, profil) = GetColumnsValues(collection);
                                    RegistryKey rk = param.RegistryKey.CreateSubKey(id);
                                    rk.SetValue(KEY_COL, cols);
                                    rk.SetValue(KEY_SIZ, sizes);
                                    rk.SetValue(KEY_PRE, previous);
                                    rk.SetValue(KEY_VIS, visible);
                                    rk.SetValue(KEY_PRO, profil);
                                }
                                #endregion

                                #region colum set and get
                                (string col, string size, string previous, string visible, string profil)
                                    GetColumnsValues(GridViewColumnCollection collection)
                                {
                                    string cols = "", sizes = "", previous = "", visible = "", profil = "";
                                    ColumnsManager cm = null;
                                    foreach (var c in collection)
                                    {
                                        cols += GetColumnID(c) + "|";
                                        sizes += c.Width.ToString(CultureInfo.InvariantCulture) + "|";
                                        if (c.Header is GridViewColumnHeaderEx ch)
                                        {
                                            previous += ch.PreviousWidth?.ToString(CultureInfo.InvariantCulture) + "|";
                                            visible += (int)ch.Visible + "|";
                                            if (cm == null) cm = ch.ColumnManager;
                                        }
                                        else
                                        {
                                            previous += "|";
                                            visible += "|";
                                        }
                                    }
                                    #region find profil
                                    if(cm != null && cm.SelectedProfil != null)
                                    {
                                        profil = cm.SelectedProfil.Value;
                                    }
                                    #endregion
#if SHOW
                                    Debug.WriteLine("--------------------------------------------");
                                    Debug.WriteLine($"colonnes = {cols}");
                                    Debug.WriteLine($"size = {sizes}");
                                    Debug.WriteLine($"previous = {previous}");
                                    Debug.WriteLine($"visible = {visible}");
                                    Debug.WriteLine($"profil = {profil}");
#endif
                                    return (cols, sizes, previous, visible, profil);
                                }

                                void SetColumnsValues(GridViewColumnCollection collection,
                                    string cols, string sizes, string previous, string visible, string profil)
                                {
                                    #region save colum & clear column in component
                                    var colsList = new List<GridViewColumn>();
                                    foreach (var item in collection)
                                    {
                                        colsList.Add(item);
                                    }
                                    #endregion

                                    #region split and test
                                    string[] colsSort = cols?.Split('|');
                                    string[] sizesWidth = sizes?.Split('|');
                                    string[] previousWidth = previous?.Split('|');
                                    string[] visibleStats = visible?.Split('|');

                                    bool isNull = colsSort == null || sizesWidth == null ||
                                        previousWidth == null || visibleStats == null;
                                    bool isSize = collection.Count == colsSort.Length - 1 &&
                                            collection.Count == sizesWidth.Length - 1 &&
                                            collection.Count == previousWidth.Length - 1 &&
                                            collection.Count == visibleStats.Length - 1;
                                    #endregion

                                    if (!isNull && isSize)
                                    {
                                        gvhrp.Columns.Clear();
                                        ColumnsManager cm = null;

                                        #region set col position & parameters
                                        for (int index = 0; index < colsSort.Length - 1; index++)
                                        {
                                            try
                                            {
                                                int pos = int.Parse(colsSort[index]);
                                                var column = colsList[pos];
                                                gvhrp.Columns.Add(column);
                                                if (!sizesWidth[index].ToLower().Equals("nan"))
                                                {
                                                    column.Width = double.Parse(sizesWidth[index], CultureInfo.InvariantCulture);
                                                }
                                                if(column.Header is GridViewColumnHeaderEx h)
                                                {
                                                    if (cm == null) cm = h.ColumnManager;
                                                    if (!string.IsNullOrEmpty(previousWidth[index]))
                                                    {
                                                        h.PreviousWidth = double.Parse(previousWidth[index], CultureInfo.InvariantCulture);
                                                    }
                                                    if (!visibleStats[index].Equals("0"))
                                                    {
                                                        var x = visibleStats[index];
                                                        Visible v = (Visible)Enum.Parse(typeof(Visible), visibleStats[index]);  
                                                        h.IsHide = (v & Visible.Hide) != 0;
                                                        h.IsHideProfil = (v & Visible.HideProfil) != 0;
                                                    }
                                                }
                                            }
                                            catch 
                                            {
                                                Debug.WriteLine("*** ERROR ******************************");
                                                // restore column if error
                                                gvhrp.Columns.Clear();
                                                foreach (var item in colsList)
                                                {
                                                    item.Width = double.NaN;
                                                    if (item.Header is GridViewColumnHeaderEx h)
                                                    {
                                                        h.IsHide = false;
                                                        h.IsHideProfil = false;
                                                    }
                                                    gvhrp.Columns.Add(item);
                                                }
                                            }
                                        }
                                        #endregion

                                        #region set profil if necessary
                                        if(cm != null && !string.IsNullOrEmpty(profil))
                                        {
                                            // chech profils exist
                                            var p = cm.Profils.FromValue(profil);
                                            if (p != null)
                                            {
                                                cm.SelectedProfil = p;
                                            }
                                        }
                                        #endregion
                                    }
                                }
                                #endregion

                                #endregion
                            }
                        };
                        #endregion
                    }
                }));
        #endregion

        #region ColumnID (private used only by SavecolumPositionID)       
        private static int GetColumnID(DependencyObject obj) => (int)obj.GetValue(ColumnIDProperty);
        private static void SetColumnID(DependencyObject obj, int value) => obj.SetValue(ColumnIDProperty, value);
        private static readonly DependencyProperty ColumnIDProperty =
            DependencyProperty.RegisterAttached("ColumnID",
                typeof(int),
                typeof(ColumnsHelper),
                new PropertyMetadata(0));
        #endregion

        #region MinColumnsWidth
        public static double? GetMinColumnsWidth(DependencyObject obj) => (double?)obj.GetValue(MinColumnsWidthProperty);
        public static void SetMinColumnsWidth(DependencyObject obj, double value) => obj.SetValue(MinColumnsWidthProperty, value);
        public static readonly DependencyProperty MinColumnsWidthProperty = DependencyProperty.RegisterAttached(
            "MinColumnsWidth",
            typeof(double?), typeof(ColumnsHelper),
            new PropertyMetadata(null, (s, e) =>
            {
                if (s is ListView listView)
                {
                    listView.Loaded += (lvs, lve) =>
                    {
                        if (listView.View is GridView view)
                        {
                            foreach (var column in view.Columns)
                            {
                                SetMinWidth(listView, column);

                                (column as INotifyPropertyChanged).PropertyChanged += (cs, ce) =>
                                {
                                    if (ce.PropertyName == nameof(GridViewColumn.ActualWidth))
                                        SetMinWidth(listView, column);
                                };
                            }
                        }
                    };
                }
            }));

        private static void SetMinWidth(ListView listView, GridViewColumn column)
        {
            double? minWidth = (double)listView.GetValue(MinColumnsWidthProperty);
            if (minWidth == null) return;
            bool canDo = true;
            if (column.Header is GridViewColumnHeaderEx gvch)
            {
                canDo = !gvch.IsHide && !gvch.IsHideProfil;
            }
            if (canDo && !double.IsNaN(column.Width) && column.ActualWidth < minWidth)
                column.Width = (double)minWidth;
        }
        #endregion

        #region STATIC
        private static T GetVisualParent<T>(DependencyObject child) where T : FrameworkElement
        {
            while ((child != null) && !(child is T))
            {
                child = VisualTreeHelper.GetParent(child);
            }
            return child as T;
        }
        private static T FindChild<T>(DependencyObject parent, string childName = null) where T : DependencyObject
        {
            if (parent == null) return null;
            T foundChild = null;
            int childrenCount = VisualTreeHelper.GetChildrenCount(parent);
            for (int i = 0; i < childrenCount; i++)
            {
                var child = VisualTreeHelper.GetChild(parent, i);
                T childType = child as T;
                if (childType == null)
                {
                    foundChild = FindChild<T>(child, childName);
                    if (foundChild != null) break;
                }
                else if (!string.IsNullOrEmpty(childName))
                {
                    var frameworkElement = child as FrameworkElement;
                    if (frameworkElement != null && frameworkElement.Name == childName)
                    {
                        foundChild = (T)child;
                        break;
                    }
                }
                else
                {
                    foundChild = (T)child;
                    break;
                }
            }
            return foundChild;
        }
        #endregion

    }
}
