﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.Windows.Media;

namespace TestGridViewColumEx
{
    public class DatePickerHelper
    {
        #region ClosePopupOnEnter              
        public static bool GetClosePopupOnEnter(DependencyObject obj) => (bool)obj.GetValue(ClosePopupOnEnterProperty);
        public static void SetClosePopupOnEnter(DependencyObject obj, bool value) => obj.SetValue(ClosePopupOnEnterProperty, value);
        public static readonly DependencyProperty ClosePopupOnEnterProperty =
            DependencyProperty.RegisterAttached("ClosePopupOnEnter",
                typeof(bool),
                typeof(DatePickerHelper),
                new PropertyMetadata(false
                , (ss, ee) =>
                {
                    if (ss is DatePicker control)
                    {
                        control.KeyUp += (ss, ee) =>
                        {
                            if (ee.Key == System.Windows.Input.Key.Enter)
                            {
                                var b = GetVisualParent<Border>(control);
                                var p = GetLogicalParent<Popup>(b);
                                p.IsOpen = false;
                            }
                        };
                    }
                }));
        #endregion

        #region SelectAllOnFocus  
        public static bool GetSelectAllOnFocus(DependencyObject obj) => (bool)obj.GetValue(SelectAllOnFocusProperty);
        public static void SetSelectAllOnFocus(DependencyObject obj, bool value) => obj.SetValue(SelectAllOnFocusProperty, value);
        public static readonly DependencyProperty SelectAllOnFocusProperty =
            DependencyProperty.RegisterAttached("SelectAllOnFocus",
                typeof(bool),
                typeof(DatePickerHelper),
                new PropertyMetadata(false
                , (ss, ee) =>
                {
                    if (ss is DatePicker control)
                    {
                        control.Loaded += delegate
                        {
                            DatePickerTextBox tb = GetChildOfType<DatePickerTextBox>(control);
                            tb?.Focus();
                            tb?.SelectAll();
                        };
                    }
                }));
        #endregion

        #region CloseOnValueChanged > close all when click on calendar           
        public static bool GetCloseOnValueChanged(DependencyObject obj) => (bool)obj.GetValue(CloseOnValueChangedProperty);
        public static void SetCloseOnValueChanged(DependencyObject obj, bool value) => obj.SetValue(CloseOnValueChangedProperty, value);
        public static readonly DependencyProperty CloseOnValueChangedProperty =
            DependencyProperty.RegisterAttached("CloseOnValueChanged",
                typeof(bool),
                typeof(DatePickerHelper),
                new PropertyMetadata(false
                , (ss, ee) =>
                {
                    if (ss is DatePicker control)
                    {
                        control.SelectedDateChanged += delegate
                        {
                            var parentPopup = GetLogicalParent<Popup>(control);
                            var insidePopup = GetChildOfType<Popup>(control);
                            if (insidePopup != null && insidePopup.IsOpen)
                            {
                                insidePopup.IsOpen = false;
                                if(parentPopup != null) parentPopup.IsOpen = false;
                            }
                            //Debug.WriteLine($"inside popup = {insidePopup}");
                            //Debug.WriteLine("CLOSE POPUP");
                        };
                    }
                }
                ));
        #endregion

        #region STATIC 
        private static T GetVisualParent<T>(DependencyObject child) where T : FrameworkElement
        {
            while ((child != null) && !(child is T))
            {
                child = VisualTreeHelper.GetParent(child);
            }
            return child as T;
        }
        private static T GetLogicalParent<T>(DependencyObject child) where T : FrameworkElement
        {
            while ((child != null) && !(child is T))
            {
                child = LogicalTreeHelper.GetParent(child);
            }
            return child as T;
        }
        private static T GetChildOfType<T>(DependencyObject depObj) where T : DependencyObject
        {
            if (depObj == null) return null;
            for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
            {
                var child = VisualTreeHelper.GetChild(depObj, i);

                var result = (child as T) ?? GetChildOfType<T>(child);
                if (result != null) return result;
            }
            return null;
        }
        #endregion
    }
}
