﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.Windows.Media;

namespace TestGridViewColumEx
{
    public class MainWindowBase : Window, INotifyPropertyChanged
    {
        #region INotifyPropertyChanged
        public event PropertyChangedEventHandler? PropertyChanged;
        protected virtual void OnPropertyChanged(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }
        #endregion

        #region Missions
        private ObservableCollection<Mission> _Missions = Mission.GetTest();
        public ObservableCollection<Mission> Missions => _Missions;
        #endregion

        #region Constructor
        public MainWindowBase()
        {
            DataContext = this;
        }
        #endregion

        #region END (COMMAND)
        RelayCommand _ENDCommand;
        public ICommand ENDCommand
        {
            get
            {
                if (_ENDCommand == null)
                {
                    _ENDCommand = new RelayCommand(param =>
                        {
                            Environment.Exit(0);
                        },
                        param => true);
                }
                return _ENDCommand;
            }
        }
        #endregion

        #region PROFILS

        #region ColumnsManager
        private ColumnsManager _ColumnsManager = null;
        public ColumnsManager ColumnsManager
        {
            get => _ColumnsManager;
            set
            {
                if (_ColumnsManager == null && value != null)
                {
                    _ColumnsManager = value;
                    PropertyChanged?.Invoke(this, new PropertyChangedEventArgs("Profils"));
                    PropertyChanged?.Invoke(this, new PropertyChangedEventArgs("SelectedProfil"));
                    _ColumnsManager.PropertyChanged += (ss, ee) =>
                    {
                        if ("SelectedProfil".Equals(ee.PropertyName))
                        {
                            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs("SelectedProfil"));
                        }
                    };
                }
            }
        }
        #endregion

        public Profil[] Profils => ColumnsManager?.Profils;

        #region SelectedProfil      
        public Profil SelectedProfil
        {
            get => ColumnsManager?.SelectedProfil;
            set
            {
                if (ColumnsManager != null)
                {
                    ColumnsManager.SelectedProfil = value;
                    OnPropertyChanged("SelectedProfil");
                }
            }
        }
        #endregion

        #region ColumnsManager (COMMAND)
        RelayCommand _ColumnsManagerCommand;
        public ICommand ColumnsManagerCommand
        {
            get
            {
                if (_ColumnsManagerCommand == null)
                {
                    _ColumnsManagerCommand = new RelayCommand(param =>
                        {
                            if (ColumnsManager == null)
                            {
                                ColumnsManager = ColumnsManager.Get(param as DependencyObject);
                            }
                        },
                        param => true);
                }
                return _ColumnsManagerCommand;
            }
        }
        #endregion

        #region ClearProfil (COMMAND)
        RelayCommand _ClearProfilCommand;
        public ICommand ClearProfilCommand
        {
            get
            {
                if (_ClearProfilCommand == null)
                {
                    _ClearProfilCommand = new RelayCommand(
                        param => SelectedProfil = null,
                        param => SelectedProfil != null);
                }
                return _ClearProfilCommand;
            }
        }
        #endregion

        #endregion

        #region STATIC
        private static T GetChildOfType<T>(DependencyObject depObj) where T : DependencyObject
        {
            if (depObj == null) return null;
            for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
            {
                var child = VisualTreeHelper.GetChild(depObj, i);

                var result = (child as T) ?? GetChildOfType<T>(child);
                if (result != null) return result;
            }
            return null;
        }
        #endregion
    }
}
